/*
 * The implementation of the Mosaic class
 * Copyright (c) 1998 - Stig Erik Sand
 *
 * $Id: mosaic.cpp,v 1.11 1998/08/01 12:43:48 stig Exp $
 */

#include "mosaic.H"
#include "tools.H"
#include "picture.H"
#include <iostream.h>
#include <qpixmap.h>
#include <qpainter.h>
#include <qimage.h>
#include <qpopmenu.h>

static QPixmap
getPixmap(const AlternativePicture& pic, int tileX, int tileY) {

    QImage *tmpPic = PictureTools::makeImage(pic.name);
    if (pic.flipped) {
//	COUTPUT("FLIPPED " << pic.name);
	tmpPic = PictureTools::makeFlip(tmpPic);
    }
    QPixmap pix;
    pix.convertFromImage(*tmpPic);
    delete tmpPic;
	
//    COUTPUT("-(Reading in: " << pic.name << ", " << pix.width() << ")");
    QWMatrix m;
    float scaleX = (float)tileX/(float)pic.width;
    float scaleY = (float)tileY/(float)pic.height;
    m.scale(scaleX, scaleY);

    QPixmap blah(pix.xForm(m));
    return blah;
}

Mosaic::Mosaic(const QImage *original, const PrefStruct& prefs)
    : indexLen(0), alternatives(NULL), pixmaps(NULL), howmany(0),
      tileX(prefs.tileSizeX), tileY(prefs.tileSizeY) {

    tiledpic = new TiledPicture(original, tileX, tileY);
//    cout << "Between" << endl;
    if (!prefs.alw_samesize) {
	irregpic = new TiledPicture(original, tileX, tileY,
				    !prefs.alw_samesize);
    }
    else {
	irregpic=NULL;
    }
//    exit(-1);
}

Mosaic::~Mosaic() {
    
    if (indexLen) {
	for (unsigned int j=0; j<indexLen; j++) delete[] picIndexes[j];
	delete[] picIndexes;
    }
    
    if (tiledpic) delete tiledpic;
    if (irregpic) delete irregpic;
    if (alternatives) delete[] alternatives;
    if (pixmaps) delete[] pixmaps;

}

int
Mosaic::getPictureVal(int tile, int alt) {
    
    return picIndexes[tile][alt];
}

QPixmap *
Mosaic::getPicture(int tile, int alt) {
    
    return &pixmaps[picIndexes[tile][alt]];
}

QPixmap
Mosaic::getColourData(const PicturePart *whichpart) {

    const Coord *cPtr;
//    const AlternativePicture *ptr = &alternatives[whichpic];

    // gambling on width being 10
    int mult = 4;
    QImage newPic(mult * 10, mult * 10, 32);

    int x=0, y=0;
    for (int i = 0; i < COORD_NUMBER; ++i) {
	cPtr = &whichpart->coords[i];
	for (int n=0; n < mult; n++) {
	    for (int m=0; m < mult; m++) {
		newPic.setPixel(x+n, y+m,
				qRgb(cPtr->colour.red, cPtr->colour.green, cPtr->colour.blue));
	    }
	}
/*	newPic.setPixel(x, y,
			qRgb(cPtr->colour.red, cPtr->colour.green, cPtr->colour.blue));
	newPic.setPixel(x+1, y,
			qRgb(cPtr->colour.red, cPtr->colour.green, cPtr->colour.blue));
	newPic.setPixel(x, y+1,
			qRgb(cPtr->colour.red, cPtr->colour.green, cPtr->colour.blue));
	newPic.setPixel(x+1, y+1,
			qRgb(cPtr->colour.red, cPtr->colour.green, cPtr->colour.blue));
*/
	x += mult;
	if (x==(mult*10)) { y+=mult; x=0; }
    }
    QPixmap tmpPic;
    tmpPic.convertFromImage(newPic);
    
    return tmpPic;
}


int
Mosaic::calculateMosaic(QPaintDevice *targetDevice,
			QPainter *pixpaint,
			QPixmap *targetPixmap,
			const PrefStruct& prefs) {

    if (!tiledpic) return -1;
//    COUTPUT("Getting parts");
    parts = tiledpic->getParts(indexLen);
    originalParts = tiledpic->getPartPixmaps();
    
    unsigned int irreglen=0;
    if (irregpic)
	irregparts = irregpic->getParts(irreglen);
    else
	irregparts = NULL;

    ASSERT(irreglen==indexLen);
    
//    COUTPUT("Len=" << indexLen << " (" << tileX << ", " << tileY << ")");
    
    COUTPUT("Building picture table..");
    alternatives = PictureTools::buildPictureTable(howmany,
						   prefs.statfilename);

    pixmaps = new QPixmap[howmany*2];
    
    picIndexes = new int*[indexLen];
    for (unsigned int k=0; k< indexLen; k++) picIndexes[k]=NULL;
    
    COUTPUT("Actually calculating (" << indexLen << ", " << howmany << ")");
    
    int dummyval=0;
    for (unsigned i=0; i < indexLen; i++) {
//	COUTPUT("Value(" << i << "): " << (int) pictureIndexes[i]);
	if (!picIndexes[i]) {
//	    COUTPUT("calculating");
	    if (irregparts && !picIndexes[i+1]) {
		picIndexes[i] = PictureTools::findMatch(alternatives,
							howmany,
							&irregparts[i],
							prefs,
							IRREGULAR);
	
	
		if (picIndexes[i][0]!=-1) {
//		    cout << "-Found a valid irregular (" << i << ", " << picIndexes[i][0] << ")" << endl;
		    int wx = irregpic->getWidth()/tileX;
		    int locy = i/wx;
		    int locx = i-wx*locy;
		    int hy = irregpic->getHeight()/tileY;
		    
		    if (locx<(wx-1) && locy<(hy-1)) {
			int *emptypicture = new int[prefs.altsize];
			int j;
			for (j=0; j<prefs.altsize; j++) emptypicture[j] = -2;
			picIndexes[i+1]=emptypicture;
			
			emptypicture = new int[prefs.altsize];
			for (j=0; j<prefs.altsize; j++) emptypicture[j] = -2;
			picIndexes[i+wx+1]=emptypicture;
			
			emptypicture = new int[prefs.altsize];
			for (j=0; j<prefs.altsize; j++) emptypicture[j] = -2;
			picIndexes[i+wx]=emptypicture;

			for (j=0; j<prefs.altsize; j++)
			    picIndexes[i][j]+=howmany;
			dummyval=1;
		    }
		    else {
//			COUTPUT("-Deleting " << i);
			delete [] picIndexes[i];
		    }
		}
	    }

	    if (!dummyval) {
		picIndexes[i] = PictureTools::findMatch(alternatives,howmany,
							&parts[i], prefs);
	    }

	    for (int n=0; n<prefs.altsize; n++) {
		if (picIndexes[i][n]>-1) {
		    if (pixmaps[picIndexes[i][n]].width()==0) {
//			COUTPUT("-Reading in picture " << picIndexes[i][0]);
			int picval = picIndexes[i][n];
			if (picval>howmany) picval-=howmany;
			pixmaps[picval] =
			    getPixmap(alternatives[picval],
				      tileX+dummyval*tileX,
				      tileY+dummyval*tileY);
		    }
		}
	    }

	    pixpaint->drawPixmap(parts[i].coords[0].x, parts[i].coords[0].y,
				 pixmaps[picIndexes[i][0]], 0, 0,
				 tileX+dummyval*tileX,
				 tileY+dummyval*tileY);
	
	
	    if (prefs.contdisp)
		bitBlt(targetDevice, MY_BORDER, MY_BORDER, targetPixmap);
	}
	dummyval=0;
    }
    cout << endl;
    
    cout << "\nUsed pictures: " << usedPictures(howmany) << " of "
	 << howmany << " pictures\n";
    cout << endl;

    return 0;
}


void
Mosaic::displayAlternative(QPainter *paint, int x, int y,
			   const int *pic) {

//    cout << "(Drawing at:" << x << "," << y << ")";
    pixmaps[*pic] = getPixmap(alternatives[*pic], tileX, tileY);
    paint->drawPixmap(x, y, pixmaps[*pic], 0, 0, tileX, tileY);

}

int
Mosaic::usedPictures(int len) const {
    
    int count=0;
    for (int i=0; i < len; i++) if (pixmaps[i].width()) count++;
    return count;

}

void
Mosaic::pasteInAlternative(QPainter *paint, int tile, int alt) {


    if (alt==0) {
	paint->drawPixmap(parts[tile].coords[0].x,
			  parts[tile].coords[0].y, 
			  originalParts[tile], 0, 0,
			  originalParts[tile].width(),
			  originalParts[tile].height());
	return;
    }
    else {
	alt--;
    }
    int picval = picIndexes[tile][alt];
//    COUTPUT("Pasting in..." << tile << " " << alt << " " << picval);
    if (picval == -2) {
	ASSERT(tile != 0);
	ASSERT(irregpic!=NULL);
	int wx = irregpic->getWidth()/tileX;

	if (picIndexes[tile-1][alt]>howmany) {
	    picval = picIndexes[tile-1][alt];
//	    COUTPUT("Left picval: " << picval);
	    tile--;
	}
	else if (picIndexes[tile-wx][alt]>howmany) {
	    picval = picIndexes[tile-wx][alt];
//	    COUTPUT("Top picval: " << picval);
	    tile-=wx;
	}
	else if (picIndexes[tile-wx-1][alt]>howmany) {
	    picval = picIndexes[tile-wx-1][alt];
//	    COUTPUT("Topleft picval: " << picval);
	    tile=tile-wx-1;
	}
	else {
	    return;
	}
    }
    else if (picval<0) return;
//    COUTPUT("Reading.. " << picval);
    QPixmap foo(pixmaps[picval]);
//    COUTPUT("Draw " << picval << " " << foo.width() << " on " << tile );
    paint->drawPixmap(parts[tile].coords[0].x,
		      parts[tile].coords[0].y, 
		      foo, 0, 0, foo.width(), foo.height());

}

void
Mosaic::buildPopup(QPopupMenu *pictlist, int whichtile, int altsize, int whichButton) {

    ASSERT(irregpic != NULL);
    ASSERT(howmany > 0);
    
    int wx = irregpic->getWidth()/tileX;
    int whichpic;
    
    if (whichButton==RightButton) {
	pictlist->insertItem(originalParts[whichtile], "Original", 0);
    }
    else /* midbutton */ {
	pictlist->insertItem(getColourData(&parts[whichtile]), "Original", 0);
    }
     
    for (int i=0; i< altsize; i++) {
//	COUTPUT("Looping " << i << " " << altsize);
	if (picIndexes[whichtile][i] > -1) {
	    whichpic = picIndexes[whichtile][i];
	}
	else {
	    ASSERT(whichtile != 0);

	    if (picIndexes[whichtile-1][i]>howmany) {
		/* left */
		whichpic = picIndexes[whichtile-1][i];
	    }
	    else if (picIndexes[whichtile-wx][i]>howmany) {
		/* top */
		whichpic = picIndexes[whichtile-wx][i];
	    }
	    else if (picIndexes[whichtile-wx-1][i]>howmany) {
		/* top-left */
		whichpic = picIndexes[whichtile-wx-1][i];
	    }
	    else {
		COUTPUT("Unknown phenomena " << whichtile );
		whichpic = -1;
	    }
	
	}
	/* actual insert */
	if (whichpic > 0) {
	    if (whichButton==RightButton) {
		pictlist->insertItem(pixmaps[whichpic], alternatives[whichpic].name, i+1);
	    }	    
	    else {
		pictlist->insertItem(getColourData(&alternatives[whichpic].data),
				     alternatives[whichpic].name, i+1);
	    }
		     
	}
    }

}
    
